#include <iostream>

//**********************************//
//RAPPEL DE C++
//**********************************//


//3. Surcharge d'opérateurs

//En C++, les classes et structs peuvent voir leur opérateurs être surchargés.
//Ce la permet, entre autres, de définir des opérations ressemblant à des formules de maths.
//On utilisera cependant ces opérateurs uniquement pour des opérations non ambigues.
// ex. vecteur1 x vecteur2 est ambigue: produit scalaire? vectoriel? terme à terme?
//     vecteur1<<vecteur2 ? concatenation, remplacement ?
//     Notez que le signe + est parfois utilisé pour la somme, parfois pour la
//  concatenation, ce qui peut également préter à des ambiguités.

class mon_vecteur
{

public:
    mon_vecteur(double _x,double _y,double _z);

    double get_x() const;
    double get_y() const;
    double get_z() const;


    //*************************************
    //opérateurs "mathématiques"
    //*************************************

    //operateurs internes:

    //surchage de l'opérateur += avec un autre vecteur
    //(notez que l'on retourne une référence de l'objet courant,
    // car on modifie la classe courant, mais pas le paramètre)
    mon_vecteur& operator+=(const mon_vecteur& vec);
    //surchage de l'opérateur += avec un scalaire
    mon_vecteur& operator+=(const double& valeur);
    //sucharge de l'opérateur -= avec un autre vecteur
    mon_vecteur& operator-=(const mon_vecteur& vec);
    //sucharge de l'opérateur -= avec un scalaire
    mon_vecteur& operator-=(double a);
    //surcharge de l'opérateur *= avec un scalaire
    mon_vecteur& operator*=(const double& value);
    //surcharge de l'opérateur /= avec un scalaire
    mon_vecteur& operator/=(const double& value);

    //operateur de négation
    //(on ne modifie pas la classe, mais on retourne une copie)
    mon_vecteur operator-() const;

    //operateurs "friend" (les friends ne sont pas hérités par les classes dérivées)

    //surchage de vec_a+vec_b: ici on ne modifie plus la classe
    //Il ne s'agit pas d'une opération de la classe à proprement parler, on le défini en friend
    friend mon_vecteur operator+(const mon_vecteur& vec_a,const mon_vecteur& vec_b);
    //surchage de scalaire+vec (nécessite une définition par friend!)
    friend mon_vecteur operator+(double a,const mon_vecteur& vec);
    //surchage de vec+scalaire
    friend mon_vecteur operator+(const mon_vecteur& vec,double a);
    //on peut faire de même avec l'operator- ...
    friend mon_vecteur operator-(const mon_vecteur& vec_a,const mon_vecteur& vec_b);
    friend mon_vecteur operator-(double a,const mon_vecteur& vec);
    friend mon_vecteur operator-(const mon_vecteur& vec,double a);

    //surchage de scalaire*vec
    friend mon_vecteur operator*(double a,const mon_vecteur& vec);
    friend mon_vecteur operator*(const mon_vecteur& vec,double a);
    //surchage de vec/scalaire (attention, pas de scalaire/vec !)
    friend mon_vecteur operator/(const mon_vecteur& vec,double a);


    //*************************************
    //opérateurs crochet et parenthèses
    //*************************************

    //On va écrire un vecteur tel que
    // "vec[0]=vec(0)=x"
    // "vec[1]=vec(1)=y"
    // "vec[2]=vec(2)=z"

    //operateur crochet: permet d'écrire: double z=vec[2];
    // Notons que l'on ne modifie pas la classe, on peut donc appeler la méthode sur un objet constant
    double operator[](unsigned int k) const;
    //Note: on aurait pu écrire aussi: const double& operator[](unsigned int k) const;

    //idem pour la parenthèse: permet d'écrire: double z=vec(2);
    double operator()(unsigned int k) const;

    //operateur crochet d'affectation: permet d'écrire vec[2]=4; (sur un objet non constant)
    double& operator[](unsigned int k); //pas de const à la fin
    //idem pour la parenthèse
    double& operator()(unsigned int k);

    //*************************************
    //opérateurs de sortie standard
    //*************************************

    // on écrit ici les données dans un flux, classiquement pour écrire avec std::cout<<
    // Ici, on a un exemple d'écriture sur une référence non constante passée en paramètre (le flux)
    friend std::ostream& operator<<(std::ostream& flux,const mon_vecteur& vec);




private:
    double x,y,z;

};

//implementation

mon_vecteur::mon_vecteur(double _x,double _y,double _z):x(_x),y(_y),z(_z){}
double mon_vecteur::get_x() const{return x;}
double mon_vecteur::get_y() const{return y;}
double mon_vecteur::get_z() const{return z;}

mon_vecteur& mon_vecteur::operator+=(const mon_vecteur& vec)
{
    x+=vec.get_x();y+=vec.get_y();z+=vec.get_z();
    return *this; //on retourne l'objet courant
}

mon_vecteur& mon_vecteur::operator+=(const double& valeur)
{x+=valeur;y+=valeur;z+=valeur;return *this;}
mon_vecteur& mon_vecteur::operator-=(const mon_vecteur& vec)
{x-=vec.get_x();y-=vec.get_y();z-=vec.get_z();return *this;}
mon_vecteur& mon_vecteur::operator-=(double a)
{return (*this)+=-a;} //on peut se servir de ce que l'on a déjà codé (simplifie la maintenance future)
mon_vecteur& mon_vecteur::operator*=(const double& value)
{x*=value;y*=value;z*=value;return *this;}
mon_vecteur& mon_vecteur::operator/=(const double& value)
{x/=value;y/=value;z/=value;return *this;}
mon_vecteur mon_vecteur::operator-() const
{return mon_vecteur(-x,-y,-z);}//notez que l'on duplique l'objet

//notez que pour les méthodes "friend", on a pas l'opérateur de résolution de portée
// dans l'implémentation (mon_vecteur::)
mon_vecteur operator+(const mon_vecteur& vec_a,const mon_vecteur& vec_b)
{
    mon_vecteur temp=vec_a; //on duplique l'objet (car on ne modifie pas l'objet courant)
    temp+=vec_b; //on se sert généralement des opérateurs "internes"
    return temp; //on retourne l'objet modifié
}

mon_vecteur operator+(double a,const mon_vecteur& vec)
{mon_vecteur temp=vec;temp+=a;return temp;}
mon_vecteur operator+(const mon_vecteur& vec,double a)
{return a+vec;}//on se sert de ce que l'on a déjà codé
mon_vecteur operator-(const mon_vecteur& vec_a,const mon_vecteur& vec_b)
{mon_vecteur temp=vec_a;temp-=vec_b;return temp;}

mon_vecteur operator-(double a,const mon_vecteur& vec)
{mon_vecteur temp=-vec;temp+=a;return temp;}//on se sert de la négation
mon_vecteur operator-(const mon_vecteur& vec,double a)
{mon_vecteur temp=vec;temp-=a;return temp;}
mon_vecteur operator*(double a,const mon_vecteur& vec)
{mon_vecteur temp=vec;temp*=a;return temp;}
mon_vecteur operator*(const mon_vecteur& vec,double a)
{return a*vec;}
mon_vecteur operator/(const mon_vecteur& vec,double a)
{return vec*(1/a);}
double mon_vecteur::operator[](unsigned int k) const
{
    if(k==0)
        return x;
    if(k==1)
        return y;
    if(k==2)
        return z;
    std::cout<<"Erreur, k="<<k<<std::endl;
    throw std::exception(); //lance une exception pouvant etre rattrapé ultérieurement
    // sans faire cracher le programme
}
double mon_vecteur::operator()(unsigned int k) const
{
    return (*this)[k];
}
double& mon_vecteur::operator[](unsigned int k)
{
    //implémentation identique à la version non const
    // (mais peut être différent dans le cas de structures dynamique
    //   ex. vecteur de taille s'adaptant automatiquement)
    if(k==0)
        return x;
    if(k==1)
        return y;
    if(k==2)
        return z;
    std::cout<<"Erreur, k="<<k<<std::endl;
    throw std::exception();
}
double& mon_vecteur::operator()(unsigned int k)
{
    return (*this)[k];
}

std::ostream& operator<<(std::ostream& flux,const mon_vecteur& vec)
{
    flux<<"("<<vec.get_x()<<","<<vec.get_y()<<","<<vec.get_z()<<")";
    return flux;
}


//utilisation
int main(int argc,char *argv[])
{
    mon_vecteur a(1,2,3);
    mon_vecteur b(4,5,6);
    mon_vecteur c=a+b; //c=(5,7,9)

    mon_vecteur d=10-c; //c=(5,3,1)
    std::cout<<d<<std::endl; //affiche c
    mon_vecteur e=d; //e=(5,3,1)
    e[0]=b(1); //e.x=b.y => e=(5,3,1);
    e+=mon_vecteur(0,1,1); //e=(5,4,2);

    std::cout<<"voici mon vecteur final: "<<e<<std::endl;

    return 0;
}
